'' Basic I2C Routines  Version 2.0
'' (c) 2014 Tracy Allen, EME Systems LLC
'' This is a modified version of BASIC_I2C_DRIVER version 1.3 by Michael Green and copyright (c) 2009
'' Terms of MIT liscense are extended.
'' Permission is given to use this in any program for the Parallax
'' Propeller processor as long as this copyright notice is included.
'' Modifications:
''    -- this special version supports clock stretching at every 0->1 transition of scl, overkill?!


''    -- The Init(scl,sda) can use any combination of pins for scl,sda
''    -- both scl and sda need pullups.  The routines manipulate dira[.] only.  Won't work with eeprom on some Parallax development boards.
''    -- negx is used as a flag to skip the address bytes, for devices that use only a device ID, no address
''    -- To handle two byte (16 bit) addresses, pass the address in as a negative value.  A positive address value is for 8 bit address
''    -- This object does not directly support addresses greater than 16 bits.  That must be handled in application code by patching the device ID.
''    -- The count parameter for strings and numbers can be passed as a negative number, for in situ small<-->large endedness conversion.
''    -- Methods for reading and writing numbers (2, 3 or 4 bytes) can likewise do msbyte<-->lsbyte inversions
''    -- sda and scl parameters are stored in eeprom
''       -- this is primarily for systems that have only one i2c bus, which may be accessed from different objects
''       -- take care only one device accesses the bus at a time;  no locks are implemented
''       -- object can be used for multiple busses so long as init[..] is called before each and every change of bus, and when the bus is not active.
''       -- method calls do not pass in the scl or sda parameters, since those are stored upon initialization
''       -- see alternate object I2C_nBus for direct support of multiple I2C busses.



CON
   ACK      = 0                        ' I2C Acknowledge
   NAK      = 1                        ' I2C No Acknowledge
   DRIVELOW = 1                        ' state of dira[.] is output to drive pin low
   PULLHIGH = 0                        ' state of dira[.] is input to allow pin to be pulled high
   LOW = 0                             ' the output register of both scl and sda will always be low, control is via dira[.] only
   NO_ADDR   = negx                     ' as a flag in the addrReg parameter position in the method call, causes method to skip sending address byte(s)

   BOOTPIN  = 28                       ' I2C Boot EEPROM SCL Pin
   EEPROM   = $A0                      ' I2C EEPROM Device Address

PUB Init(_scl,_sda)                               ' pass in both pins, call to set up the buss or to change from one buss to another
  sda := _sda
  scl := _scl
  dira[scl] := PULLHIGH                                 ' both scl and sda are pulled high by resistor
  dira[sda] := PULLHIGH
  outa[scl] := LOW                                      ' leave a zero in the scl and sda output latches.
  outa[sda] := LOW                                      ' so turning pin to output will drive the pin low.

PUB Twiddle                                             ' some devices can use this to get out of a bad state.
   repeat 9
      outa[scl] := DRIVELOW                             ' Put out up to 9 clock pulses
      outa[scl] := PULLHIGH
      if ina[sda]                                       ' Repeat if SDA not driven high
         quit
   return (ina[scl] << 1) + ina[sda]

PUB Start                                               ' SDA goes HIGH to LOW with SCL HIGH
  dira[scl] := PULLHIGH
  if ina[scl] == 0
    hold
  dira[sda] := PULLHIGH
  dira[sda] := DRIVELOW
  dira[scl] := DRIVELOW

PUB Stop                                                ' SDA goes LOW to HIGH with SCL High
  dira[scl] := PULLHIGH
  if ina[scl] == 0
    hold
  dira[sda] := PULLHIGH

PUB Write(data) : ackbit
'' Write i2c data.  Data byte is output MSB first, SDA data line is valid
'' only while the SCL line is HIGH.  SDA can change only when SCL is low.
'' Data is always 8 bits (+ ACK/NAK).
'' SCL and SDA are both left in the LOW state.
'' returns 0 ACK, 1 if NAK
  ackbit := 0
  data <<= 24
  repeat 8
    dira[sda] := !(data <-= 1)
    dira[scl] := PULLHIGH
    if ina[scl] == 0
      hold
    dira[scl] := DRIVELOW
  dira[sda] := PULLHIGH
  dira[scl] := PULLHIGH
  if ina[scl] == 0
    hold
  ackbit := ina[sda]
  dira[scl] := DRIVELOW
  dira[sda] := DRIVELOW


PUB Read(ackbit): data
'' Read in i2c data, Data byte is input MSB first, SDA data line is
'' valid only while the SCL line is HIGH, and changes only when SCL is low
'' SCL and SDA left in LOW state.
'' returns byte value
  dira[sda] := PULLHIGH
  repeat 8
    dira[scl] := PULLHIGH
    if ina[scl] == 0
      hold
    data := (data << 1) | ina[sda]
    dira[scl] := DRIVELOW
  dira[sda] := !ackbit
  dira[scl] := PULLHIGH
  if ina[scl] == 0
    hold
  dira[scl] := DRIVELOW
  dira[sda] := DRIVELOW

PUB hold

  waitpne(0, |<scl, 0)



PUB ReadPage(devSel, addrReg, dataPtr, count) : ackbit
'' Read in a counted block of i2c data.  Device select code is devSel, 8 bit form.
'  Device starting address is addrReg. Data address is dataPtr.  Number of bytes is count.
'' addrReg is either positive, 8 bits, or make addrReg negative to specify a two-byte address.
'' Use addrReg=negx as a flag to skip the device address preamble and go right into the read operation.
'' For eeproms > 64 kbytes, must handle 17th and higher bits ad hoc to merge with devSel
'' The data is returned in the byte array pointed to at entry by dataPtr.
'' The function call return 1 if all acks are okay, 0 otherwise, ACKS/NAKS come only from writing devSel and addrReg, not from reading data
'' ----------
  Start                       '
  ackbit := Write(devSel)
  if addrReg > negx           ' negx as flag causes routine to skip the address altogether, direct to data bytes
    if addrReg < 0             ' use negative addrReg value to invoke 18 bit address
      ackbit |= Write(||addrReg >> 8 & $FF)              ' msb
    ackbit |= Write(||addrReg & $FF)                    ' lsb
  'waitpne(0, |<SCL, 0)
  ' waitcnt(clkfreq/100+cnt)
  Start                          ' Reselect the device for reading
  ackbit |= Write(devSel | 1)  ' 8 bit address, 1 in lsb for read operation
  if count < 0
    || count
    dataPtr += count - 1
    repeat count - 1
      byte[dataPtr--] := Read(ACK)
  else
    repeat count - 1
      byte[dataPtr++] := Read(ACK)
  byte[dataPtr] := Read(NAK)
  Stop
  ' return 0 if all ACKs okay, 1 if one or more NAKs


PUB ReadByte(devSel, addrReg)
'' reads one byte from device.
'' positive value in addrReg sends 8 bit address, negative value in addrReg sends 16 bit address
'' negx in addrReg skips sending any address bytes  (devSel only)
   ReadPage(devSel, addrReg, @result, 1)

PUB ReadNumber(devSel, addrReg, count)
'' count is length of the number in bytes (1,2,3,4), to be transferred from device to Prop.
'' positive count transfers number byte for byte, negative count swaps the endedness, that is, msb from device goes to lsb in Prop.
'' positive value in addrReg sends 8 bit address, negative value in addrReg sends 16 bit address
'' negx in addrReg skips sending any address bytes  (devSel only)
   ReadPage(devSel, addrReg, @result, count)

PUB WritePage(devSel, addrReg, dataPtr, count) : ackbit
'' Write out a block of i2c data.  Device select code is devSel, 8 bit form.
'  Device starting address is addrReg. Data address is dataPtr.  Number of bytes is count.
'' Use negative addrReg to specify two byte address.
'' Use addrReg=negx to specify null addrReg, goes right into the data bytes
'' Use count=0 to skip sending any data (i.e. devSel and addrReg only.)
'' For eeproms > 64 kbytes, must handle 17th and higher bits ad hoc to merge with devSel
'' Return zero if all acks are okay, zero otherwise
'' ----------
  Start
  ackbit := Write(devSel)   '  devSel is 8 bit, zero in lsb
  if addrReg > negx         ' negx flag causes skip direct to sending bytes
    if addrReg<0             ' negative addrReg signals 2 byte address (eeprom)
      ackbit |= Write(||addrReg >> 8 & $FF)   ' msb
    ackbit |= Write(||addrReg & $FF)           ' lsb, or the only address byte
  if count<0
    || count
    dataptr += count
    repeat count
      ackbit |= Write(byte[--dataPtr])
  else
    repeat count
      ackbit |= Write(byte[dataPtr++])      '' write successive bytes small endian like prop.
    ackbit |= Write(byte[dataPtr++])
  Stop
  ' return 0 if all ACKs okay, 1 if one or more NAKs

PUB WriteByte(devSel, addrReg, datum)
'' writes one byte to device.
'' positive value in addrReg sends 8 bit address, negative value in addrReg sends 16 bit address
'' negx in addrReg skips sending any address bytes  (devSel only)
'' function call returns 0 if all ACKs are okay, 1 if one or more NAKs
   return  ReadPage(devSel, addrReg, @datum, 1)

PUB WriteNumber(devSel, addrReg, datum, count)
'' count is length of the number in bytes, to be transferred from Prop to device.
'' positive count transfers number byte for byte, negative count swaps the endedness, that is, msb from prop goes to lsb in device.
'' positive value in addrReg sends 8 bit address, negative value in addrReg sends 16 bit address
'' negx in addrReg skips sending any address bytes  (devSel only)
'' function call returns 0 if all ACKs are okay, 1 if one or more NAKs
   return  ReadPage(devSel, addrReg, @datum, count)

PUB Hello(devSel) : ackbit
'' Wait for a previous write to complete.  Device select code is devSel.  Device
'' starting address is addrReg.  The device will not respond if it is busy.
'' function call returns 0 if device at devSel responded (ACK, present and ready), otherwise 1 (NAK, not present, or currently busy)
   Start
     ackbit := Write(devSel)
   Stop

DAT
  scl byte 28
  sda byte 29
  lockID byte 0
  zzzz byte 0
  timeout long 8_000_000   ' 0.1 second at 80 MHz, for clock stretching clkfreq/10000 * hold
  instance byte 1

DAT

{{
┌──────────────────────────────────────────────────────────────────────────────────────┐
│                           TERMS OF USE: MIT License                                  │
├──────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this  │
│software and associated documentation files (the "Software"), to deal in the Software │
│without restriction, including without limitation the rights to use, copy, modify,    │
│merge, publish, distribute, sublicense, and/or sell copies of the Software, and to    │
│permit persons to whom the Software is furnished to do so, subject to the following   │
│conditions:                                                                           │
│                                                                                      │
│The above copyright notice and this permission notice shall be included in all copies │
│or substantial portions of the Software.                                              │
│                                                                                      │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,   │
│INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A         │
│PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT    │
│HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION     │
│OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE        │
│SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                                │
└──────────────────────────────────────────────────────────────────────────────────────┘
}}

